/*
firetalk.c - FireTalk wrapper definitions
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#define FIRETALK

#include "toc.h"
#include "firetalk.h"
#include "firetalk-int.h"

typedef void (*ptrtotoc)(void *, ...);

/* Global variables */
int firetalkerror;
static struct s_firetalk_handle *handle_head = NULL;
const char *firetalkerrors[] = {
/*  0 */	"Success",
/*  1 */	"Connection failed",
/*  2 */	"Usernames do not match",
/*  3 */	"Packet transfer error",
/*  4 */	"Invalid username or password",
/*  5 */	"Invalid sequence number from server",
/*  6 */	"Invalid frame type from server",
/*  7 */	"Packet too long",
/*  8 */	"Server problem; try again later",
/*  9 */	"Unknown error",
/* 10 */	"You are blocked",
/* 11 */	"Unknown packet received from server",
/* 12 */	"Invalid callback number",
/* 13 */	"Invalid username",
/* 14 */	"Username not found in list",
/* 15 */	"Server disconnected",
/* 16 */	"Unable to create socket",
/* 17 */	"Unable to resolve hostname",
/* 18 */	"Wrong server version",
/* 19 */	"User is currently unavailable",
/* 20 */	"User information is currently unavilable",
/* 21 */	"You are sending messages too fast; last message was dropped",
/* 22 */	"Chat room is currently unavailable",
/* 23 */	"Incoming message delivery failure",
/* 24 */	"User disconnected",
/* 25 */	"Server response was formatted incorrectly",
/* 26 */  "You have requested idle to be reset too fast",
/* 27 */  "Invalid room name",
/* 28 */  "Invalid message (too long?)",
/* 29 */  "Invalid protocol"
};

const char *firetalkprotocols[] = {
/*  0 */  "AOL Instant Messenger/TOC"
};

/* Internal function definitions */

/* firetalk_find_by_toc searches the firetalk handle list for the toc handle passed, and returns the firetalk handle */
struct s_firetalk_handle *firetalk_find_handle(const int protocol, void * const c) {
	static struct s_firetalk_handle *iter;
	static int i;
	iter = handle_head;
	while (iter) {
		for (i = 0; i < FIRETALK_MAX_CONNECTIONS; i++)
			if (iter->handles[protocol][i] == c)
				return iter;
		iter = iter->next;
	}
	return NULL;
}

int firetalk_find_index(const struct s_firetalk_handle * const conn, const int protocol, void * const c) {
	static int i;

	for (i = 0; i < FIRETALK_MAX_CONNECTIONS; i++)
		if (conn->handles[protocol][i] == c)
			return i;

	return -1;
}

int firetalk_im_internal_add_buddy(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_buddy *iter;
	static struct s_firetalk_handle *conn;
	int (*comparenicks)(const char * const, const char * const);

	conn = c;

	switch (protocol) {
		case FP_AIMTOC:
			comparenicks = toc_compare_nicks;
			break;
		default:
			firetalkerror = FE_BADPROTO;
			return FE_BADPROTO;
			break;
	}

	iter = conn->buddy_head[protocol][connection];

	while (iter) {
		if (!comparenicks(iter->nickname,nickname)) {
			firetalkerror = FE_SUCCESS;
			return FE_SUCCESS; /* not an error, user is in buddy list */
		}
		iter = iter->next;
	}

	iter = conn->buddy_head[protocol][connection];
	conn->buddy_head[protocol][connection] = malloc(sizeof(struct s_firetalk_buddy));
	if (!conn->buddy_head[protocol][connection]) {
		perror("malloc");
		exit(1);
	}
	conn->buddy_head[protocol][connection]->next = iter;
	conn->buddy_head[protocol][connection]->nickname = malloc(strlen(nickname)+1);
	if (!conn->buddy_head[protocol][connection]->nickname) {
		perror("malloc");
		exit(1);
	}
	strcpy(conn->buddy_head[protocol][connection]->nickname,nickname);
	conn->buddy_head[protocol][connection]->online = 0;
  conn->buddy_head[protocol][connection]->away = 0;
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int firetalk_im_internal_add_deny(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_deny *iter;
	static struct s_firetalk_handle *conn;
	int (*comparenicks)(const char * const, const char * const);

	conn = c;

	switch (protocol) {
		case FP_AIMTOC:
			comparenicks = toc_compare_nicks;
			break;
		default:
			firetalkerror = FE_BADPROTO;
			return FE_BADPROTO;
			break;
	}

	iter = conn->deny_head[protocol][connection];

	while (iter) {
		if (!comparenicks(iter->nickname,nickname)) {
			firetalkerror = FE_SUCCESS;
			return FE_SUCCESS; /* not an error, user is in buddy list */
		}
		iter = iter->next;
	}

	iter = conn->deny_head[protocol][connection];
	conn->deny_head[protocol][connection] = malloc(sizeof(struct s_firetalk_deny));
	if (!conn->deny_head[protocol][connection]) {
		perror("malloc");
		exit(1);
	}
	conn->deny_head[protocol][connection]->next = iter;
	conn->deny_head[protocol][connection]->nickname = malloc(strlen(nickname)+1);
	if (!conn->deny_head[protocol][connection]->nickname) {
		perror("malloc");
		exit(1);
	}
	strcpy(conn->deny_head[protocol][connection]->nickname,nickname);
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int firetalk_im_internal_remove_buddy(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_buddy *iter;
	static struct s_firetalk_buddy *iter2;
	static struct s_firetalk_handle *conn;
	int (*comparenicks)(const char * const, const char * const);

	conn = c;

	switch (protocol) {
		case FP_AIMTOC:
			comparenicks = toc_compare_nicks;
			break;
		default:
			firetalkerror = FE_BADPROTO;
			return FE_BADPROTO;
			break;
	}

	iter2 = NULL;
	iter = conn->buddy_head[protocol][connection];
	while (iter) {
		if (!comparenicks(nickname,iter->nickname)) {
			if (iter2)
				iter2->next = iter->next;
			else
				conn->buddy_head[protocol][connection] = iter->next;
			free(iter);
			firetalkerror = FE_SUCCESS;
			return FE_SUCCESS;
		}
		iter2 = iter;
		iter = iter->next;
	}

	firetalkerror = FE_NOTFOUND;
	return FE_NOTFOUND;
}

int firetalk_im_internal_remove_deny(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_deny *iter;
	static struct s_firetalk_deny *iter2;
	static struct s_firetalk_handle *conn;
	int (*comparenicks)(const char * const, const char * const);

	conn = c;

	switch (protocol) {
		case FP_AIMTOC:
			comparenicks = toc_compare_nicks;
			break;
		default:
			firetalkerror = FE_BADPROTO;
			return FE_BADPROTO;
			break;
	}

	iter2 = NULL;
	iter = conn->deny_head[protocol][connection];
	while (iter) {
		if (!comparenicks(nickname,iter->nickname)) {
			if (iter2)
				iter2->next = iter->next;
			else
				conn->deny_head[protocol][connection] = iter->next;
			free(iter);
			firetalkerror = FE_SUCCESS;
			return FE_SUCCESS;
		}
		iter2 = iter;
		iter = iter->next;
	}

	firetalkerror = FE_NOTFOUND;
	return FE_NOTFOUND;
}

void firetalk_callback_im_getmessage(const int protocol, void * const c, const char * const sender, const int automessage, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IM_GETMESSAGE])
		conn->callbacks[FC_IM_GETMESSAGE](conn,protocol,firetalk_find_index(conn,protocol,c),sender,automessage,message);
}

void firetalk_callback_im_buddyonline(const int protocol, void * const c, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IM_BUDDYONLINE])
		conn->callbacks[FC_IM_BUDDYONLINE](conn,protocol,firetalk_find_index(conn,protocol,c),nickname);
}

void firetalk_callback_im_buddyoffline(const int protocol, void * const c, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IM_BUDDYOFFLINE])
		conn->callbacks[FC_IM_BUDDYOFFLINE](conn,protocol,firetalk_find_index(conn,protocol,c),nickname);
}

void firetalk_callback_im_buddyaway(const int protocol, void * const c,  const char * const nickname) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IM_BUDDYAWAY])
		conn->callbacks[FC_IM_BUDDYAWAY](conn,protocol,firetalk_find_index(conn,protocol,c),nickname);
}

void firetalk_callback_im_buddyunaway(const int protocol, void * const c, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IM_BUDDYUNAWAY])
		conn->callbacks[FC_IM_BUDDYUNAWAY](conn,protocol,firetalk_find_index(conn,protocol,c),nickname);
}

void firetalk_callback_error(const int protocol, void * const c, const int error, const char * const roomoruser) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_ERROR])
		conn->callbacks[FC_ERROR](conn,protocol,firetalk_find_index(conn,protocol,c),error,roomoruser);
}

void firetalk_callback_disconnect(const int protocol, void * const c, const int error) {
	static struct s_firetalk_handle *conn;
	static struct s_firetalk_buddy *buddyiter;
	static struct s_firetalk_buddy *buddyiter2;
	static struct s_firetalk_deny *denyiter;
	static struct s_firetalk_deny *denyiter2;
	static int i;

	conn = firetalk_find_handle(protocol,c);
	i = firetalk_find_index(conn,protocol,c);

	conn->handles[FP_AIMTOC][i] = NULL;

	buddyiter = conn->buddy_head[FP_AIMTOC][i];
	while (buddyiter) {
		buddyiter2 = buddyiter;
		buddyiter = buddyiter->next;
		free(buddyiter2);
	}
	conn->buddy_head[FP_AIMTOC][i] = NULL;

	denyiter = conn->deny_head[FP_AIMTOC][i];
	while (denyiter) {
		denyiter2 = denyiter;
		denyiter = denyiter->next;
		free(denyiter2);
	}
	conn->deny_head[FP_AIMTOC][i] = NULL;

	if (conn->callbacks[FC_DISCONNECT])
		conn->callbacks[FC_DISCONNECT](conn,protocol,i,error);
}

void firetalk_callback_gotinfo(const int protocol, void * const c, const char * const nickname, const char * const info, const int warning, const int idle) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_GOTINFO])
		conn->callbacks[FC_GOTINFO](conn,protocol,firetalk_find_index(conn,protocol,c),nickname,info,warning,idle);
}

void firetalk_callback_idleinfo(const int protocol, void * const c, char const * const nickname, const long idletime) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_IDLEINFO])
		conn->callbacks[FC_IDLEINFO](conn,protocol,firetalk_find_index(conn,protocol,c),nickname,idletime);
}

void firetalk_callback_doinit(const int protocol, void * const c) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_DOINIT])
		conn->callbacks[FC_DOINIT](conn,protocol,firetalk_find_index(conn,protocol,c));
}

void firetalk_callback_setidle(const int protocol, void * const c, long * const idle) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_SETIDLE])
		conn->callbacks[FC_SETIDLE](conn,protocol,firetalk_find_index(conn,protocol,c),idle);
}

void firetalk_callback_eviled(const int protocol, void * const c, const int newevil, const char * const eviler) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_EVILED])
		conn->callbacks[FC_EVILED](conn,protocol,firetalk_find_index(conn,protocol,c),newevil,eviler);
}

void firetalk_callback_chat_joined(const int protocol, void * const c, const char * const room) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_JOINED])
		conn->callbacks[FC_CHAT_JOINED](conn,protocol,firetalk_find_index(conn,protocol,c),room);
}

void firetalk_callback_chat_left(const int protocol, void * const c, const char * const room) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_LEFT])
		conn->callbacks[FC_CHAT_LEFT](conn,protocol,firetalk_find_index(conn,protocol,c),room);
}

void firetalk_callback_chat_getmessage(const int protocol, void * const c, const char * const room, const char * const from, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_GETMESSAGE])
		conn->callbacks[FC_CHAT_GETMESSAGE](conn,protocol,firetalk_find_index(conn,protocol,c),room,from,message);
}

void firetalk_callback_chat_getwhisper(const int protocol, void * const c, const char * const room, const char * const from, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_GETWHISPER])
		conn->callbacks[FC_CHAT_GETWHISPER](conn,protocol,firetalk_find_index(conn,protocol,c),room,from,message);
}

void firetalk_callback_chat_invited(const int protocol, void * const c, const char * const room, const char * const from, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_INVITED])
		conn->callbacks[FC_CHAT_INVITED](conn,protocol,firetalk_find_index(conn,protocol,c),room,from,message);
}

void firetalk_callback_chat_user_joined(const int protocol, void * const c, const char * const room, const char * const who) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_USER_JOINED])
		conn->callbacks[FC_CHAT_USER_JOINED](conn,protocol,firetalk_find_index(conn,protocol,c),room,who);
}

void firetalk_callback_chat_user_left(const int protocol, void * const c, const char * const room, const char * const who) {
	static struct s_firetalk_handle *conn;
	conn = firetalk_find_handle(protocol,c);
	if (conn->callbacks[FC_CHAT_USER_LEFT])
		conn->callbacks[FC_CHAT_USER_LEFT](conn,protocol,firetalk_find_index(conn,protocol,c),room,who);
}

/* External function definitions */
void *firetalk_create_handle() {
	static struct s_firetalk_handle *c;
	static int i;
	static int j;
	c = handle_head;
	handle_head = malloc(sizeof(struct s_firetalk_handle));
	handle_head->prev = NULL;
	handle_head->next = c;
	for (i = 0; i < FP_MAX; i++)
		for (j = 0; j < FIRETALK_MAX_CONNECTIONS; j++) {
			handle_head->handles[i][j] = NULL;
			handle_head->buddy_head[i][j] = NULL;
			handle_head->deny_head[i][j] = NULL;
		}
	return handle_head;
}

void firetalk_destroy_handle(void * const c) {
	static struct s_firetalk_handle *conn;
	static int i;
	conn = c;
	if (conn->prev)
		conn->prev->next = conn->next;
	else
		handle_head = conn->next;
	if (conn->next)
		conn->next->prev = conn->prev;
	for (i = 0; i < FIRETALK_MAX_CONNECTIONS; i++)
		if (conn->handles[FP_AIMTOC][i])
			toc_disconnect(conn->handles[FP_AIMTOC][i]);
	free(conn);
}

void firetalk_perror(const char * const string) {
	fprintf(stderr,"%s: %s\n",string,firetalkerrors[firetalkerror]);
}

int firetalk_disconnect(void * const c, const int protocol, const int connection) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC) {
		conn->handles[FP_AIMTOC][connection] = NULL;
		return toc_disconnect(conn->handles[FP_AIMTOC][connection]);
	}
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_signon(void * const c, const int protocol, const int connection, const char * const server, const int port, const char * const username, const char * const password) {
	static struct s_firetalk_handle *conn;
	static int i;
	conn = c;
	if (protocol == FP_AIMTOC) {
		if (! (conn->handles[FP_AIMTOC][connection] = toc_connect_server(server,port)))
			return FE_CONNECT;
		/* register ourselves for the disconnect event so we can clean up */
		if ((i = toc_signon(conn->handles[FP_AIMTOC][connection],username,password))) {
			conn->handles[FP_AIMTOC][connection] = NULL;
			firetalkerror = i;
			return i;
		}
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int firetalk_register_callback(void * const c, const int type, void (*function)(void *, const int, const int, ...)) {
	static struct s_firetalk_handle *conn;
	conn = c;
	/* here is a translation table from FIRETALK to the sub modules */
	if (type < 0 || type > FC_MAX) {
		firetalkerror = FE_CALLBACKNUM;
		return FE_CALLBACKNUM;
	}
	conn->callbacks[type] = function;
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int firetalk_im_add_buddy(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	static int ret;
	conn = c;
	ret = FE_BADPROTO;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		ret = toc_im_add_buddy(conn->handles[FP_AIMTOC][connection],nickname);

	if (ret != FE_SUCCESS) {
		firetalkerror = ret;
		return ret;
	}
	
	return firetalk_im_internal_add_buddy(conn,protocol,connection,nickname);
}

int firetalk_im_remove_buddy(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	static int ret;
	conn = c;
	ret = FE_BADPROTO;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		ret = toc_im_remove_buddy(conn->handles[FP_AIMTOC][connection],nickname);

	if (ret != FE_SUCCESS) {
		firetalkerror = ret;
		return ret;
	}

	return firetalk_im_internal_remove_buddy(conn,protocol,connection,nickname);
}

int firetalk_im_add_deny(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	static int ret;
	conn = c;
	ret = FE_BADPROTO;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		ret = toc_im_add_deny(conn->handles[FP_AIMTOC][connection],nickname);

	if (ret != FE_SUCCESS) {
		firetalkerror = ret;
		return ret;
	}

	return firetalk_im_internal_add_deny(conn,protocol,connection,nickname);
}

int firetalk_im_remove_deny(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	static int ret;
	conn = c;
	ret = FE_BADPROTO;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		ret = toc_im_remove_deny(conn->handles[FP_AIMTOC][connection],nickname);

	if (ret != FE_SUCCESS) {
		firetalkerror = ret;
		return ret;
	}

	return firetalk_im_internal_add_deny(conn,protocol,connection,nickname);
}

int firetalk_save_config(void * const c, const int protocol, const int connection) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_save_config(conn->handles[FP_AIMTOC][connection]);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_im_upload_buddies(void * const c, const int protocol, const int connection) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_im_upload_buddies(conn->handles[FP_AIMTOC][connection]);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_im_upload_denies(void * const c, const int protocol, const int connection) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_im_upload_denies(conn->handles[FP_AIMTOC][connection]);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_im_send_message(void * const c, const int protocol, const int connection, const char * const dest, const char * const message, const int auto_flag) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_im_send_message(conn->handles[FP_AIMTOC][connection],dest,message,auto_flag);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_get_info(void * const c, const int protocol, const int connection, const char * const nickname) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_get_info(conn->handles[FP_AIMTOC][connection],nickname);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_set_info(void * const c, const int protocol, const int connection, const char * const info) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_set_info(conn->handles[FP_AIMTOC][connection],info);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_im_list_buddies(void * const c, const int protocol, const int connection) {
	static struct s_firetalk_handle *conn;
	static struct s_firetalk_buddy *buddyiter;
	conn = c;

	if (protocol < 0 || protocol >= FP_MAX) {
		firetalkerror = FE_BADPROTO;
		return FE_BADPROTO;
	}

	buddyiter = conn->buddy_head[protocol][connection];
	while (buddyiter) {
		if (conn->callbacks[FC_IM_LISTBUDDY])
			conn->callbacks[FC_IM_LISTBUDDY](conn,protocol,connection,buddyiter->nickname,buddyiter->online,buddyiter->away);
		buddyiter = buddyiter->next;
	}

	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int firetalk_set_away(void * const c, const int protocol, const int connection, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_set_away(conn->handles[FP_AIMTOC][connection],message);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_im_evil(void * const c, const int protocol, const int connection, const char * const who) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_im_evil(conn->handles[FP_AIMTOC][connection],who);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_chat_join(void * const c, const int protocol, const int connection, const char * const room) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_chat_join(conn->handles[FP_AIMTOC][connection],room);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_chat_part(void * const c, const int protocol, const int connection, const char * const room) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_chat_part(conn->handles[FP_AIMTOC][connection],room);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_chat_send_message(void * const c, const int protocol, const int connection, const char * const room, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_chat_send_message(conn->handles[FP_AIMTOC][connection],room,message);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_chat_invite(void * const c, const int protocol, const int connection, const char * const room, const char * const who, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_chat_invite(conn->handles[FP_AIMTOC][connection],room,who,message);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_chat_whisper(void * const c, const int protocol, const int connection, const char * const room, const char * const who, const char * const message) {
	static struct s_firetalk_handle *conn;
	conn = c;
	if (protocol == FP_AIMTOC && conn->handles[FP_AIMTOC][connection])
		return toc_chat_whisper(conn->handles[FP_AIMTOC][connection],room,who,message);
	firetalkerror = FE_BADPROTO;
	return FE_BADPROTO;
}

int firetalk_select(void * const c) {
	return firetalk_select_custom(c,0,NULL,NULL,NULL,NULL);
}

int firetalk_select_custom(void * const c, int n, fd_set *fd_read, fd_set *fd_write, fd_set *fd_except, struct timeval *timeout) {
	static struct s_firetalk_handle *conn;
	static int ret;
	static int tempsock;
	static fd_set *my_read;
	static fd_set *my_write;
	static fd_set *my_except;
	static fd_set internal_read;
	static fd_set internal_write;
	static fd_set internal_except;
	static struct timeval internal_timeout;
	static struct timeval *my_timeout;
	static int i;

	conn = c;

	my_read = fd_read;
	my_write = fd_write;
	my_except = fd_except;
	my_timeout = timeout;

	if (!my_read) {
		my_read = &internal_read;
		FD_ZERO(my_read);
	}

	if (!my_write) {
		my_write = &internal_write;
		FD_ZERO(my_write);
	}

	if (!my_except) {
		my_except = &internal_except;
		FD_ZERO(my_except);
	}

	if (!my_timeout) {
		my_timeout = &internal_timeout;
		my_timeout->tv_sec = 60;
		my_timeout->tv_usec = 0;
	}

	if (my_timeout->tv_sec > 60)
		my_timeout->tv_sec = 60;

	for (i = 0; i < FIRETALK_MAX_CONNECTIONS; i++) {
		if (conn->handles[FP_AIMTOC][i]) {
			tempsock = toc_get_fd(conn->handles[FP_AIMTOC][i]);
			if (tempsock >= n)
				n = tempsock + 1;
			FD_SET(tempsock,my_read);
			FD_SET(tempsock,my_except);
			toc_set_idle(conn->handles[FP_AIMTOC][i]);
		}
	}

	ret = select(n,my_read,my_write,my_except,my_timeout);

	for (i = 0; i < FIRETALK_MAX_CONNECTIONS; i++) {
		if (conn->handles[FP_AIMTOC][i]) {
			tempsock = toc_get_fd(conn->handles[FP_AIMTOC][i]);
			if (FD_ISSET(tempsock,my_read)) {
				ret--;
				toc_got_data(conn->handles[FP_AIMTOC][i]);
			}
			if (FD_ISSET(tempsock,my_except)) {
				ret--;
				toc_disconnect(conn->handles[FP_AIMTOC][i]);
			}
		}
	}

	return ret;
}
